/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001, 2002 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
/**
An indexed collection stores references to objects.  It supports
indexed access to objects (via the [] operator), addition of
new objects (with automatics memory management), and removal
of objects from the collection.

<p>New elements are appended to the end of the indexed collection.
When an element is removed, the last element of the collection fills
the vacant slot, thereby avoiding holes in the indexing.

<p>Virtual functions do not make calls to other virtual functions
*/
class indexed: public object {
	public:		
		indexed();	
		/**
		Capacity is the initial memory set aside to hold elements.
		When the collection becomes full, its size is increased by grow
		slots.
		*/
		indexed(unsigned int capacity, unsigned int grow = DEFAULT_CONTAINER_GROWTH);
		/**
		Pointers to the objects in the original index are copied into the new index,
		the objects themselves are not copied.
		*/
		indexed(const indexed &src);
		/// The last index is at size - 1
		inline unsigned int getSize() const { return size; }
		/// Reports the current capacity of the data structure
		inline unsigned int getCapacity() const { return capacity; }
		/**
		Reduces the capacity to fit its current size.
		Undefined if size == 0.
		*/
		void trim();
		/// Increases the capacity by adjustment value.
		void enlarge(unsigned adjustment);
		/// Returns a pointer to the object stored at index
		inline object* operator[](unsigned index) const {
			return table[index];
		}
		/// Returns a pointer to the object stored at index
		inline object* getObj(unsigned index) const {
			return table[index];
		}
		/// Swaps the pointers at index i and j
		virtual void swap(unsigned int i, unsigned int j);
		/// Appends the items in src to this indexed.
		virtual void append(const indexed& src);
		/// Deletes all objects contained in the indexed. Sets size to zero.
		virtual void empty();
		/// Removes all objects from container.
		virtual void removeAll();
		/**
		Adds an object to the end of the indexed (last index).
		*/
		virtual void add(object *obj);
		/// Removes and returns the object at the indicated index.
		virtual object *remove(unsigned int index);
		/// Deletes the structure.  Has no effect on elements.
		virtual ~indexed();
	private:	
		unsigned int capacity;
		unsigned int grow;
		unsigned int size;
		object** table;
		bst duptree;
};

/// The original bag has been replaced by a typedef on the indexed class
typedef indexed bag;
