/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <string.h>
#include <adevs/adevs.h>
using namespace adevs;

// a port-value pair for receiving-sending events.
// NULL bag version for comparing with others in set.

class _pv_pair: public object {

private:	port_t port;
		bag* value;

public:		inline _pv_pair (port_t port, bool _bag = true):object () { 
			this->port = port;
		     	if (_bag) value = new bag ();
			else value = NULL;
			}
		inline unsigned getSize () const { return value->getSize (); }
		inline port_t getPort () const { return port; }
		inline void clearPort () { value->removeAll (); }
		inline bag* getValues () { return value; }
		inline void addValue (object* obj) { value->add (obj); }
		bool operator== (const object& obj) const;
		bool operator!= (const object& obj) const;
		unsigned hashcode () const { return port; }
		~_pv_pair ();

};

bool _pv_pair::operator== (const object& obj) const {

// Unsafe cast, but since this class is only used internal to the devs object
// we can ensure that it is used safely by using only _pv_pair objects in
// our containers (i.e. no mixing of object types within containers)
const _pv_pair* p = (_pv_pair*)&obj;
return p->port == port;

}

bool _pv_pair::operator!= (const object& obj) const {

return !((*this) == obj);

}

_pv_pair::~_pv_pair () { 

if (value != NULL) delete value; 

}

// end of class _pv_pair

devs::devs (devs::ModelType type):object () {

typeCode = type;
t = tN = tL = 0;
adjust_outports = false;
outport_tabs = NULL;
isThreaded = true;

}

devs::devs (const char* name, devs::ModelType type):object () {

typeCode = type;
t = tN = tL = 0;
adjust_outports = false;
this->name = name;
outport_tabs = NULL;
isThreaded = true;

}

void devs::hold (stime_t dt) {

// Passivate if dt is INFINITY
if (dt == INFINITY) tN = INFINITY;
// Otherwise hold for dt time units
else tN = t + dt;

}

bag* devs::inputOnPort (port_t inport) const {

_pv_pair port (inport, false);

if (inports.instanceCount (port) == 0) {
   UndefinedException err (this, inport);
   throw err;
   }

_pv_pair* pv = (_pv_pair*)inports[inports.indexOf (port)];
return pv->getValues ();

}

void devs::output (port_t outport, object* value, bool gc) {

_pv_pair port (outport, false);

if (outports.instanceCount (port) == 0) {
   UndefinedException err (this, outport);
   throw err;
   }

_pv_pair* pv = (_pv_pair*)outports[outports.indexOf (port)];
pv->addValue (value);

if (gc) garbage.add (value);

}

void devs::addInputPort (port_t inport) {

_pv_pair port (inport, false);

if (inports.instanceCount (port) == 0) 
   inports.add (new _pv_pair (inport));

}

void devs::addOutputPort (port_t outport) {

_pv_pair port (outport, false);

if (outports.instanceCount (port) == 0) {
   outports.add (new _pv_pair (outport));
   adjust_outports = true;
   }

}

bool devs::hasInputPort (port_t inport) const {

_pv_pair port (inport, false);
return inports.instanceCount (port) != 0; 

}

bool devs::hasOutputPort (port_t outport) const {

_pv_pair port (outport, false);
return outports.instanceCount (port) != 0; 

}

void devs::_reset (stime_t t0) {

tL = t = t0;
tN = INFINITY;
clearPorts ();

}

const port_t* devs::outputPorts (int *count) {

if (adjust_outports) {
   outport_tabs = (port_t *)realloc (outport_tabs, 
   outports.getSize () * sizeof (port_t));
   for (unsigned i = 0; i < outports.getSize (); i++) {
      _pv_pair* pv = (_pv_pair*)outports[i]; 
      outport_tabs[i] = pv->getPort ();
      }
   adjust_outports = false;
   } 

*count = outports.getSize ();
return outport_tabs;

}

bag* devs::outputOnPort (port_t outport) const {

_pv_pair port (outport, false);

if (outports.instanceCount (port) == 0) {
   UndefinedException err (this, outport);
   throw err;
   }

_pv_pair* pv = (_pv_pair*)outports[outports.indexOf (port)];
return pv->getValues ();

}

void devs::clearPorts (indexed& io_garbage) {

io_garbage.append (garbage);
garbage.removeAll ();
clearPorts ();

}

void devs::clearPorts () {

for (unsigned i = 0; i < outports.getSize (); i++) {
   _pv_pair* pv = (_pv_pair*)outports[i];
   pv->clearPort(); 
   }
for (unsigned i = 0; i < inports.getSize (); i++) {
   _pv_pair* pv = (_pv_pair*)inports[i];
   pv->clearPort(); 
   }
// Failure to empty the garbage can result in a potential
// memory leak, but will prevent failures of other models
// trying to remove garbage that has been deleted.
// This would only occur in the case of a fatal exception
// in any event (i.e. cycle does not complete), 
// so we won't worry about it here.

}

void devs::input (port_t inport, object* value) {

_pv_pair port (inport, false);

if (inports.instanceCount (port) == 0) {
   UndefinedException err (this, inport);
   throw err;
   }

_pv_pair* pv = (_pv_pair*)inports[inports.indexOf(port)];
pv->getValues()->add(value);

}

devs::~devs () {

if (outport_tabs != NULL) free (outport_tabs);
clearPorts ();
outports.empty ();
inports.empty ();

}
