#include <adevs/adevs.h>
#include <algorithm> 
#include <math.h>
#include <iostream>
using namespace adevs;

// Radius of the Earth (meters)
static const double r = 1738E3;
// Gravitational constant
static const double c2 = 4.925E12;
// Thrust at t < tdec
static const double f1 = 36350.0;
// Thrust at t >= tdec
static const double f2 = 1308.0;
// Fuel efficiency for thrust f1
static const double c11 = 0.000277;
// Fuel efficiency for thrust f2
static const double c12 = 0.000277;
// Initial height above surface of the moon
static const double h0 = 59404.0;
// Initial velocity
static const double v0 = -2003.0;
// Initial mass
static const double m0 = 1038.358;
// Simulation time limit
static const double tmx = 230.0;
// Time to fire the motor for deceleration
static const double tdec = 43.2;
// Time to shut off the motor
static const double tend = 210.0;
// Quantum size
static const double D = 0.1;
// Simulation clock
static double t;

// Unit step function
static double step (double x) {

if (t >= x) return 1.0;
return 0.0;

}

int main () {

// Thrust provided by the lander motor
double thrust;
// Altitude of the lander
double h;
// Lander mass and its first derivative
double m, mdot;
// Lander velocity
double v;
// Fuel efficiency
double c1;
// Total acceleration and acceleration due to gravity
double a, g;
// Time of last event
double t_last;
// Time remaining until motor activate and shutdown
double tend_s, tdec_s;
// Time step for the current iteration
double dt;
// State vector (h,v,m) where
// h is the altitude of the lander
// v is the velocity of the lander
// m is the mass of the lander
qode q(3);
q.setParams(D);
q.init (0, h0);
q.init (1, v0);
q.init (2, m0);
h = q.qstate(0);
v = q.qstate(1);
m = q.qstate(2);
// Derivatives
double dq [3];

// Initial time,time of last event, and time advance
t = t_last = dt = 0.0;
// Set time remaining until motor events
tend_s = tend; 
tdec_s = tdec;

// Output file header (gnuplot comment)
std::cout << "#t h v thrust m g " << std::endl;
// While above the surface and not going up
while (t <= tmx && h >= 0.0 && v <= 0.0) {
   // Compute next system state
   g = c2 / pow ((h+r), 2.0);
   thrust = (1.0 - step (tend)) * (f1 - (f1 - f2) * step (tdec));
   c1 = (1.0 - step (tend)) * (c11 - (c11 - c12) * step (tdec));
   mdot = -c1 * fabs (thrust);
   a = (1.0 / m) * (thrust - m * g);
	dq[0] = v;
	dq[1] = a;
	dq[2] = mdot;
   q.integ (dt, dq);
   // Compute next event time
   tend_s -= dt;
   tdec_s -= dt;
   if (tend_s < EPSILON) tend_s = INFINITY;
   if (tdec_s < EPSILON) tdec_s = INFINITY;
   dt = std::min(std::min(tend_s, tdec_s),q.ta());
	// Get value of state variables at time t + dt
	h = q.out(0,dt);
	v = q.out(1,dt);
	m = q.out(2,dt);
	// Update simulation clock
   t += dt;
   // Record the current state every 0.02 time units
   if (t - t_last >= 0.02) {
      t_last = t;
      std::cout << t << " " << h  << " " << v << " " << thrust << " " << m << " " << g << std::endl;
      }
   }

}
