/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "xblock.h"
#include <iostream>
#include <stdlib.h>
using namespace adevs;

// input and output ports
const port_t block::status = 1;
const port_t block::ball_collision = 2;
const port_t block::paddle_collision = 3;
const port_t block::paddle_motion = 4;
const port_t block::ball_motion = 5;

block::block ():atomic ("block") {

fixed = false;
addOutputPort (ball_collision);
addOutputPort (paddle_collision);
addOutputPort (status);
addInputPort (ball_motion);
addInputPort (paddle_motion);

}

block::block (const rectangle& geom_init):
atomic ("block") {

fixed = false;
this->geom_init = geom = geom_init;

addOutputPort (ball_collision);
addOutputPort (paddle_collision);
addOutputPort (status);
addInputPort (ball_motion);
addInputPort (paddle_motion);

}

block::block (const rectangle& geom_init, eStatus stat):
atomic ("block") {

fixed = true;
this->stat = stat_init = stat;
this->geom_init = geom = geom_init;

addOutputPort (ball_collision);
addOutputPort (paddle_collision);
addOutputPort (status);
addInputPort (ball_motion);
addInputPort (paddle_motion);

}

void block::reset () {

// need to report our status at the beginning.  Afterwards,
// only report if block is destroyed.
report_status = true;

// clean out the collision lists
ball_collisions.removeAll ();
paddle_collisions.removeAll ();

geom = geom_init;
if (fixed) stat = stat_init;

// if status is not fixed, select OK or INVULNERABLE at random
else if (random () < (RAND_MAX / 4)) stat = INVULNERABLE;
else stat = OK;
hold (0);

}

void block::delta_int () {

// clear the report_status falg
report_status = false;

// clear out the collision lists
ball_collisions.removeAll ();
paddle_collisions.removeAll ();

passivate ();

}

void block::delta_ext () {

// if block is destroyed, passivate and don't respond to any more
// events.
if (stat == DESTROYED) {
   passivate ();
   return;
   }

// check for a collision with a ball 
bag* values = inputOnPort (ball_motion);
for (unsigned i = 0; i < values->getSize (); i++) {
   motionEvent* e = (motionEvent*)values->getObj (i);
   circle* ball_geom = (circle*)e->getPosition ();
   // if a collision is detected, add a collision event to the 
   // ball collision list and (if appropriate) set the status
   // of the block to DESTROYED.
   if (geom.overlaps (*ball_geom)) {
      ball_collisions.add (new collisionEvent (e->getSource(), &geom)); 
      if (stat != INVULNERABLE) {
         report_status = true;
         stat = DESTROYED;
         }
      }
   }

// check for a collision with a paddle
values = inputOnPort (paddle_motion);
for (unsigned i = 0; i < values->getSize (); i++) {
   motionEvent* e = (motionEvent*)values->getObj (i);
   // if a collision is detected, add a collision event
   // to the paddle collision list
   if (geom.overlaps (*(rectangle*)e->getPosition ())) {
      paddle_collisions.add (new collisionEvent (e->getSource(), &geom)); 
      }
   }

// generate output immediately
hold (0);

}

void block::output_func () {

// report the status if it has changed or on a reset
if (report_status)
   output (status, new statusEvent (this, stat, &geom));

// output all of the detected collisions
for (unsigned i = 0; i < ball_collisions.getSize(); i++) 
   output (ball_collision, ball_collisions[i]); 
for (unsigned i = 0; i < paddle_collisions.getSize(); i++) 
   output (paddle_collision, paddle_collisions[i]); 

}

void block::delta_conf () {

delta_int ();
delta_ext ();

}

block::~block () { }
