/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "xblock.h"
#include <iostream>
using namespace adevs;

// input and output ports
const port_t paddle::collision = 0;
const port_t paddle::vcontrol = 1;
const port_t paddle::ball_motion = 2;
const port_t paddle::motion = 3;
const port_t paddle::ball_collision = 4;
const port_t paddle::stop = 5;

// motion resolution (distance considered significant)
const double paddle::MOTION_RES = 0.5;

// zero velocity vector
const vector paddle::V_ZERO (0, 0);

paddle::paddle ():atomic ("paddle") {

v_paddle.x = 1.5;
setupPorts ();

}

paddle::paddle (const rectangle& geom_init):atomic ("paddle") {

v_paddle.x = 1.5;
this->geom_init = geom = geom_op = geom_init;
setupPorts ();

}

paddle::paddle (const vector& v_paddle):atomic ("paddle") {

this->v_paddle = v_paddle;
setupPorts ();

}

paddle::paddle (const rectangle& geom_init, const vector& v_paddle):atomic ("paddle") {

this->geom_init = geom = geom_op = geom_init;
this->v_paddle = v_paddle;
setupPorts ();

}

void paddle::setupPorts () {

addInputPort (ball_motion);
addInputPort (vcontrol);
addInputPort (collision);
addInputPort (stop);
addOutputPort (ball_collision);
addOutputPort (motion);

}

void paddle::reset () {

collisions.empty ();
geom = geom_op = geom_init;
v = V_ZERO;
vector dp = V_ZERO;
hold (0);

}

void paddle::delta_int () {

// clear the collision list
collisions.removeAll ();

// update the current position of the paddle
geom = geom_op;
// compute the next position of the paddle
vector dp = distInTime (timeToDist (MOTION_RES, v), v);
geom_op.deltaPos (dp);
// hold until the next position is reached
hold (timeToDist (MOTION_RES, v));

}

void paddle::delta_ext () {

// update the position of the paddle to reflect the elapsed time
geom.deltaPos (distInTime (elapsed (), v));
geom_op = geom;


// passivate if a stop message is received. 
bag* values = inputOnPort (stop);
if (values->getSize () > 0) {
   passivate ();
   return;
   }

bool collision_occuring = false;

// check for collisions with a wall
values = inputOnPort (collision);
if (values->getSize () > 0) { 
   collision_occuring = true;
   collisionEvent* e = (collisionEvent*)values->getObj (0);
   const rectangle* r = (rectangle*)e->getGeometry ();
   // back out of the collision area
   vector vback;
   if (r->getX () <= geom.getX ()) vback = v_paddle;
   else vback = -1 * v_paddle;
   while (geom.overlaps (*r)) {
      geom.deltaPos (vback);
      }
   geom_op = geom;
   // stop the paddle
   v = V_ZERO;
   hold (0);
   }

// check for control signals (skip this if stopping because of a collision)
values = inputOnPort (vcontrol);
if (values->getSize () > 0 && !collision_occuring) {
   // update the paddle velocity 
   paddleControlEvent* e = (paddleControlEvent*)values->getObj (0);
   if (e->getControl () == LEFT) v = v_paddle * -1;
   else if (e->getControl () == RIGHT) v = v_paddle;
   else v = V_ZERO;
   hold (0);
   }

// check for ball collision
values = inputOnPort (ball_motion);
for (unsigned i = 0; i < values->getSize (); i++) {
   motionEvent* e = (motionEvent*)values->getObj (i);
   // if a collision occured, add the collision event to the collision list
   if (geom.overlaps (*(circle*)e->getPosition ())) {
      hold (0);
      collisions.add (new collisionEvent (e->getSource(), &geom)); 
      }
   }

// unless holding for an immediate output, compute next position
// and hold for the time required to reach it.
if (ta () > 0) {
   dp = distInTime (sigma(), v);
   geom_op.deltaPos (dp);
   hold (sigma ());
   }

}

void paddle::delta_conf () {

delta_int ();
delta_ext ();

}

void paddle::output_func () {

// output new position
output (motion, new motionEvent (this, &geom));

// output all collision events
for (unsigned i = 0; i < collisions.getSize(); i++) 
   output (ball_collision, collisions[i]); 

}

paddle::~paddle () { }
