/**********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "xblock.h"
#include <iostream.h>
#include <math.h>
#include <stdlib.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <stdio.h>
using namespace adevs;

// class for holding old positions
class __pos: public object {

public:
	int x, y, width, height;
	bool first;
	inline __pos ():object () {
		first = true;
		x = y = width = height = 0;
        	}
	inline ~__pos () { }

};

// input and output ports
const port_t viewer::ball_motion = 0;
const port_t viewer::block_position = 1;
const port_t viewer::paddle_motion = 2;
const port_t viewer::paddle_control = 3;
const port_t viewer::ball_status = 4;
const port_t viewer::stop = 5;

// period with which the input queue is checked
const stime_t viewer::PERIOD = 1.0;

/**
The code here is standard for X windows applications.
Does some parsing of the command line looking for display
options.
*/

viewer::viewer (char** argv, int argc, int width, int height):
atomic ("viewer") {

addInputPort (ball_motion);
addInputPort (block_position);
addInputPort (paddle_motion);
addInputPort (ball_status);
addOutputPort (paddle_control);
addOutputPort (stop);

// open the display
display = XOpenDisplay (NULL);

if (display == NULL) {
   cout << "xblock: Could not connect to display." << endl;
   exit (-1);
   }

screen = DefaultScreen (display);

// get screen dimensions

// create the window at (0,0)

win = XCreateSimpleWindow (display, XRootWindow (display, screen),
			0, 0, width, height, 0, 
			XBlackPixel (display, screen),
			XBlackPixel (display, screen));
   
// set the size hints.  Defaults for all other properties.

size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
size_hints.x = size_hints.y = 0;
size_hints.width = size_hints.min_width = size_hints.max_width = width;
size_hints.height = size_hints.min_height = size_hints.max_height = height;

XSetWMProperties (display, win, NULL, NULL, argv, argc,
			&size_hints, NULL, NULL);


memset (&event_mask, 0xff, sizeof (long));
XSelectInput (display, win, KeyPressMask | StructureNotifyMask | KeyReleaseMask);

// xor graphics contexts.

gc = XCreateGC (display, XRootWindow (display, screen), 0, NULL);

XSetFunction (display, gc, GXxor);
XSetForeground (display, gc, XWhitePixel (display, screen));
XSetBackground (display, gc, XBlackPixel (display, screen));

// make the window visible

XMapWindow (display, win);
XClearWindow (display, win);
XAutoRepeatOff (display);

// initial number of balls and the score

balls = 2;
score = 0;

}

void viewer::reset () {

quit = false;
block_count = 0;
XClearWindow (display, win);
checkEvents ();
XSync (display, False);

if (control_event != NULL) delete control_event;
control_event = NULL;
pos_history.clear ();
drawScore (score, true);
drawBallCount (balls, true);
hold (PERIOD);

}

void viewer::delta_int () {

// clear the last control output
control_event = NULL;

// if game is over, passivate
if (quit || (block_count == 0 && timeCurrent () > PERIOD))
   passivate ();
// otherwise hold until the next user input period
else {
   hold (PERIOD);
   }

// check the event queue
checkEvents ();

}

void viewer::delta_ext () {

// clear and then draw the ball position
bag* values = inputOnPort (ball_motion);
for (unsigned i = 0; i < values->getSize(); i++) {
   motionEvent* e = (motionEvent*)values->getObj (i);
   const circle* c = (circle*)e->getPosition ();
   __pos* old_pos = (__pos*)pos_history.find (*e->getSource ());
   if (old_pos == NULL) {
      old_pos = new __pos ();
      pos_history.insert (const_cast<object*>(e->getSource ()), old_pos);
      }
   drawBall (c->getX (), c->getY (), c->getRadius (), old_pos);
   }

// clear and then draw the block position
values = inputOnPort (block_position);
for (unsigned i = 0; i < values->getSize(); i++) {
   statusEvent* e = (statusEvent*)values->getObj (i);
   const rectangle* r = (rectangle*)e->getGeometry ();
   // update the score and remaining block count if the block was destroyed
   if (e->getStatus () == DESTROYED) {
      score++;
      block_count--;
      drawScore (score);
      }
   // otherwise, if block is destroyable increment the block count
   else if (e->getStatus () == OK) {
      block_count++;
      }
   drawBlock (r->getX (), r->getY (), r->getWidth (), r->getHeight (),
		e->getStatus ());
   }

// clear and the draw the paddle
values = inputOnPort (paddle_motion);
for (unsigned i = 0; i < values->getSize (); i++) {
   motionEvent* e = (motionEvent*)values->getObj (i);
   const rectangle* r = (rectangle*)e->getPosition ();
   __pos* old_pos = (__pos*)pos_history.find (*e->getSource ());
   if (old_pos == NULL) {
      old_pos = new __pos ();
      pos_history.insert (const_cast<object*>(e->getSource ()), old_pos);
      }
   drawPaddle (r->getX (), r->getY (), r->getWidth (), r->getHeight (), old_pos);
   }

// check the ball status
values = inputOnPort (ball_status);
for (unsigned i = 0; i < values->getSize (); i++) {
   statusEvent* e = (statusEvent*)values->getObj (i);
   // if the ball is destroyed, decrement the ball count.
   // if it is zero, end the game
   if (e->getStatus () == DESTROYED) {
      balls--;
      drawBallCount (balls);
      if (balls < 0) {
         quit = true;
         hold (0);
         }
      }
   }

}

void viewer::delta_conf () { delta_int (); delta_ext (); }

void viewer::output_func () { 

// output the paddle control event
if (control_event != NULL) {
   output (paddle_control, control_event);
   }

// output a stop event if the game is over
if (quit || (block_count == 0 && timeCurrent () > PERIOD))
   output (stop, new intg ());

}
  
viewer::~viewer () {

XFreeGC (display, gc);
XAutoRepeatOn (display);
XCloseDisplay (display);
pos_history.clear ();

}

void viewer::drawScore (int score, bool first) {

if (!first) 
   XDrawString (display, win, gc, 10, 10, last_score, strlen (last_score));

sprintf (last_score, "Score: %d", score);
XDrawString (display, win, gc, 10, 10, last_score, strlen (last_score));

}
 
void viewer::drawBallCount (int balls, bool first) {

if (!first) 
   XDrawString (display, win, gc, 10, 25, last_ball_count, strlen (last_ball_count));

sprintf (last_ball_count, "Balls: %d", balls);
XDrawString (display, win, gc, 10, 25, last_ball_count, strlen (last_ball_count));

}

void viewer::drawPaddle (double x, double y, double width, double height,
			__pos* old) {

int _x = (int)rint (x);
int _y = (int)rint(y);
int _w = (int)rint(width);
int _h = (int)rint(height);

if (!old->first) 
   XDrawRectangle (display, win, gc, old->x, old->y, old->width, old->height);
else old->first = false;

old->x = _x;
old->y = _y;
old->width = _w;
old->height = _h;

XDrawRectangle (display, win, gc, _x, _y, _w, _h);

}

void viewer::drawBall (double x, double y, double radius, __pos* old) {

int _x = (int)rint (x - radius / 2.0);
int _y = (int)rint (y - radius / 2.0); 
int _radius = (int)rint (radius * 2.0); 

if (!old->first) 
   XDrawArc (display, win, gc, old->x, old->y, old->width, old->height, 
		0, 360*64);
else old->first = false;

old->x = _x;
old->y = _y;
old->width = old->height = _radius;

XDrawArc (display, win, gc, _x, _y, _radius, _radius, 0, 360*64);

}

void viewer::drawBlock (double x, double y, double width, double height,
		eStatus status) {

int _x = (int)rint (x);
int _y = (int)rint (y);
int _w = (int)rint (width);
int _h = (int)rint (height);

XDrawRectangle (display, win, gc, _x, _y, _w, _h);

if (status != INVULNERABLE)
   XFillRectangle (display, win, gc, _x, _y, _w, _h);

}

void viewer::checkEvents () { 

int vk;

while (XCheckMaskEvent (display, event_mask, &event) == True) {
   if (event.type == KeyPress) {
      vk = XKeycodeToKeysym (display, event.xkey.keycode, 0);
      if (control_event != NULL) delete control_event;
      if (vk == XK_Right) control_event = new paddleControlEvent (RIGHT);
      else if (vk == XK_Left) control_event = new paddleControlEvent (LEFT);
      else if (vk == XK_Q || vk == XK_q) {
         quit = true;
         }
      }
   else if (event.type == KeyRelease) {
      vk = XKeycodeToKeysym (display, event.xkey.keycode, 0);
      if (control_event != NULL) delete control_event;
      if (vk == XK_Right || vk == XK_Left) 
         control_event = new paddleControlEvent (STOP);
      }
   } 
}

