/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
#include "point.h"
#include "viewer.h"
#include "cell.h"
#include <assert.h>
using namespace adevs;

// Define the cell's ports.
const port_t cell::u = 0;
const port_t cell::n = 1;
const port_t cell::q = 2;
const port_t cell::b = 3;

// All cells share a common display object.  This is set externally.
viewer* cell::view = NULL;
// Mutex to synchronize access to the viewer
Mutex cell::viewer_mutex;

cell::cell (const point& location, bool alive)
:atomic () {

loc = location;
this->alive = alive;
nalive = 0;
addInputPort (n);
addOutputPort (u);
addOutputPort (q);
addOutputPort (b);

}

/**
Reset and initialization function.
*/
void cell::reset () {

// If we are a new cell with > 0 neighbors, decide if we are alive
if (nalive > 0) alive = nalive == 3;
// if we are alive, then should output our state at the next timestep
if (alive) hold (EPSILON);
// otherwise, we're dead so just sit there.
else passivate ();

}

/**
Internal transition function.
*/
void cell::delta_int () {

// assume we are going to passivate.  Can reschedule later if necessary.
passivate ();

// if we are alive but doomed due to the number of living neighbors
// then change phase to dead and hold (hold time in this case was chosen
// arbitrarily)
if (alive && !canLive ()) {
   alive = false; 
   hold (EPSILON);
   }

}

/**
External transition function.
*/
void cell::delta_ext () {

// get the bag of messages on the input port.
bag* values = inputOnPort (n);

// for each message in the bag, increment our neighbor
// count by the value of the message (+1 if a neighbor
// was born, -1 if just deceased).
for (unsigned i = 0; i < values->getSize (); i++) {
   intg* x = (intg*)values->getObj (i);
   nalive += x->value;
   }

// if we have exactly three neighbors and are currently dead, then we are born.
// Hold so that we can inform our neighbors of our new state.
if (nalive == 3 && !alive) {
   alive = true;
   hold (EPSILON);
   }

// otherwise, if we are doomed due to our neighbor count,
// set our phase to dead and hold.
else if (!canLive () && alive) {
   alive = false;
   hold (EPSILON);
   }

// otherwise, just sit there.
else 
   passivate ();

}

/**
Confluent transition function.
*/
void cell::delta_conf () {

delta_ext ();

}

/**
Output function.
*/
void cell::output_func () {

// tell the viewer our current state.
viewer_mutex.lock ();
view->setState (loc, alive);
viewer_mutex.unlock ();
// if we are alive, output a 1 and a birth message
if (alive) {
   assert (nalive == 3 || nalive == 0);
   output (u, new intg (1));
   output (b, loc);
   }
// else output a -1 and our position
else output (u, new intg (-1));
// if we have zero neighbors and are not alive, output our position
if (!alive && nalive == 0) output (q, loc);

}

cell::~cell () { }
